
#include <cassert>
#include <iostream>
#include <unordered_set>

#include <ASTNodes.h>
#include <SemanticAnalysis.h>
#include <SemanticException.h>
#include <StringConstants.h>

using namespace lcpl;

SemanticAnalysis::SemanticAnalysis(Program *p)
    : program(p), typeTable(p), symbolTable() {
  assert(program && "Expected non-null program");
}

void SemanticAnalysis::runAnalysis() {
  assert(program && "Expected non-null program");
  visit(program);
}

bool SemanticAnalysis::visit(Program *p) {
  checkMainClassAndMethod();
  checkInheritanceGraph();

  std::map<Class *, bool> processed;
  for (auto c : *p) {
    if (processed[c]) {
      throw DuplicateClassException(c);
    }

    processed[c] = true;

    if (!visit(c)) {
      return false;
    }
  }

  return true;
}

bool SemanticAnalysis::visit(Class *c) {
  if (typeTable.isBuiltinClass(c)) {
    return true;
  }

  // create a scope and add all attributes to it
  SymbolTable::Scope classScope(symbolTable);
  for (auto currentClass = c; currentClass;
       currentClass = typeTable.getParentClass(currentClass)) {
    for (auto f : *currentClass) {
      if (f->isAttribute()) {
        auto attr = static_cast<Attribute *>(f);
        symbolTable.insert(attr);

        auto attrType = typeTable.getType(attr->getType());
        typeTable.setType(attr, attrType);
      }
    }
  }

  // add self to symbol table
  std::unique_ptr<LocalDefinition> self(
      new LocalDefinition(0, "self", c->getName()));
  symbolTable.insert(self.get());
  typeTable.setType(self.get(), typeTable.getType(c));

  checkFeatures(c);

  return ASTVisitor::visit(c);
}

void SemanticAnalysis::checkFeatures(Class *c) {
  for (auto f : *c) {
    if (f->isAttribute()) {
      auto attr = static_cast<Attribute *>(f);
      if (typeTable.getAttribute(typeTable.getParentClass(c), f->getName())) {
        throw DuplicateAttrException(attr, c);
      }
    } else {
      assert(f->isMethod() && "Unknown feature kind");
      auto method = static_cast<Method *>(f);
      auto hidden =
          typeTable.getMethod(typeTable.getParentClass(c), f->getName());

      if (hidden) {
        // check that the signatures match
        if (method->getReturnType() != hidden->getReturnType()) {
          throw InvalidMethodSignatureException(c, method);
        }

        auto methodB = method->begin(), methodE = method->end();
        auto hiddenB = hidden->begin(), hiddenE = hidden->end();

        for (; methodB != methodE && hiddenB != hiddenE; ++methodB, ++hiddenB) {
          auto paramType = typeTable.getType((*methodB)->getType());
          auto hiddenParamType = typeTable.getType((*hiddenB)->getType());

          if (paramType != hiddenParamType) {
            throw InvalidMethodSignatureException(c, method);
          }
        }

        if (methodB != methodE || hiddenB != hiddenE) {
          throw InvalidMethodSignatureException(c, method);
        }
      }
    }
  }
}

bool SemanticAnalysis::visit(Feature *f) { return ASTVisitor::visit(f); }

bool SemanticAnalysis::visit(Attribute *a) {
  auto attrType = typeTable.getType(a->getType());

  auto init = a->getInit();
  if (!init) {
    return true;
  }

  if (!visit(init)) {
    return false;
  }

  // TODO: Check type compatibility, throw WrongTypeException if anything wrong

  return true;
}

void SemanticAnalysis::checkMainClassAndMethod() {
	// TODO: check if the main class is ok, throw MethodNotFoundException is not
}

void SemanticAnalysis::checkInheritanceGraph() {
	// TODO: throw BadInheritanceException if the inheritance graph is incorrect
}

bool SemanticAnalysis::visit(Method *m) {
  std::unordered_set<std::string> paramNames;

  SymbolTable::Scope methodScope(symbolTable);
  for (auto param : *m) {

  	// TODO: look for duplicate parameters and throw DuplicateParamException
    paramNames.insert(param->getName());

    if (!visit(param)) {
      return false;
    }
  }

  auto body = m->getBody();
  if (body) {
    if (!visit(body)) {
      return false;
    }

    // TODO: check for return type compatibility and throw WrongTypeException
  }

  return true;
}

bool SemanticAnalysis::visit(FormalParam *param) {
  symbolTable.insert(param);
  typeTable.setType(param, typeTable.getType(param->getType()));
  return true;
}

bool SemanticAnalysis::visit(Expression *e) { return ASTVisitor::visit(e); }

bool SemanticAnalysis::visit(IntConstant *ic) {
  typeTable.setType(ic, typeTable.getIntType());
  return true;
}

bool SemanticAnalysis::visit(StringConstant *sc) {
  typeTable.setType(sc, typeTable.getStringType());
  return true;
}

bool SemanticAnalysis::visit(NullConstant *nc) {
  typeTable.setType(nc, typeTable.getNullType());
  return true;
}

bool SemanticAnalysis::visit(Symbol *s) {
  auto who = symbolTable.lookup(s->getName());
  typeTable.setType(s, typeTable.getType(who));

  return true;
}

bool SemanticAnalysis::visit(Block *b) {
  SymbolTable::Scope blockScope(symbolTable);

  // visit all expressions first
  if (!ASTVisitor::visit(b)) {
    return false;
  }

  // set the type to match the type of the last expression
  if (b->begin() != b->end()) {
    typeTable.setType(b, typeTable.getType(b->back()));
  } else {
    typeTable.setType(b, typeTable.getVoidType());
  }

  return true;
}

bool SemanticAnalysis::visit(Assignment *a) {
  auto rhs = a->getExpression();

  if (!rhs) {
    throw MissingOperandException(a);
  }

  if (!visit(rhs)) {
    return false;
  }

  // TODO: the setType is wrong! Also throw IncompatibleOperandsException if the operands don't match
  typeTable.setType(a, typeTable.getVoidType());

  return true;
}

bool SemanticAnalysis::visit(BinaryOperator *bo) {
	auto lhs = bo->getLHS();
	auto rhs = bo->getRHS();

	if (!lhs || !rhs) {
		throw MissingOperandException(bo);
	}

	if (!visit(lhs) && !visit(rhs))
	{
		return false;
	}

	// TODO: This is wrong! check for type compatiblity and set the right type!
	// throw WrongTypeException or IncompatibleOperandsException if anything wrong

	typeTable.setType(bo, typeTable.getIntType());

  return true;
}

bool SemanticAnalysis::visit(UnaryOperator *uo) {
  auto operand = uo->getOperand();

  if (!operand) {
    throw MissingOperandException(uo);
  }

  assert(uo->getOperatorKind() != UnaryOperator::Invalid &&
         "Can't have invalid unary operators");

  if (!visit(operand)) {
    return false;
  }

  if (typeTable.getType(operand) != typeTable.getIntType()) {
    throw WrongTypeException(typeTable.getType(operand), typeTable.getIntType(),
                             uo);
  }

  typeTable.setType(uo, typeTable.getIntType());

  return true;
}

bool SemanticAnalysis::visit(Cast *c) {
  auto toCast = c->getExpressionToCast();

  if (!toCast) {
    throw MissingOperandException(c);
  }

  if (!visit(toCast)) {
    return false;
  }

  typeTable.setType(c, typeTable.getType(c->getType()));

  return true;
}

bool SemanticAnalysis::visit(Substring *s) {
  auto str = s->getString();
  auto start = s->getStart();
  auto end = s->getEnd();

  if (!str || !start || !end) {
    throw MissingOperandException(s);
  }

  if (!visit(str) || !visit(start) || !visit(end)) {
    return false;
  }

  if (typeTable.getType(str) != typeTable.getStringType()) {
    throw WrongTypeException(typeTable.getType(str), typeTable.getStringType(),
                             s);
  }

  if (typeTable.getType(start) != typeTable.getIntType()) {
    throw WrongTypeException(typeTable.getType(start), typeTable.getIntType(),
                             s);
  }

  if (typeTable.getType(end) != typeTable.getIntType()) {
    throw WrongTypeException(typeTable.getType(end), typeTable.getIntType(), s);
  }

  typeTable.setType(s, typeTable.getStringType());

  return true;
}

bool SemanticAnalysis::visit(Dispatch *d) {
  auto obj = d->getObject();

  if (obj) {
    if (!visit(obj)) {
      return false;
    }
  } else {
    auto self = symbolTable.lookup("self");
    obj = dynamic_cast<Expression *>(self);
    assert(obj && "Invalid lookup");
  }

  auto objType = typeTable.getType(obj);
  auto objClass = objType->getClass();
  if (!objClass) {
    throw DispatchOnInvalidObjException(d->getName(), objType);
  }

  auto method = typeTable.getMethod(objClass, d->getName());
  if (!method) {
    throw MethodNotFoundException(d->getName(), objClass);
  }

  // TODO: detect the type that should be returned by the dispatch
  // check the number and type of the arguments
  // throw TooManyArgsException, NotEnoughArgsException
  // or WrongTypeException according to the error type

  auto paramIt = method->begin();
  for (auto arg : *d) {
    if (!visit(arg)) {
      return false;
    }
  }

  typeTable.setType(d, typeTable.getVoidType());

  return true;
}

bool SemanticAnalysis::visit(StaticDispatch *d) {
  auto obj = d->getObject();

  if (obj) {
    if (!visit(obj)) {
      return false;
    }
  } else {
    auto self = symbolTable.lookup("self");
    obj = dynamic_cast<Expression *>(self);
    assert(obj && "Invalid lookup");
  }

  // TODO:
  // detect which method should be called
  // visit all parameters of that methos
  // detect the type that should be returned by the dispatch
  // throw DispatchOnInvalidObjException , MethodNotFoundException, TooManyArgsException, NotEnoughArgsException
  // or WrongTypeException according to the error type

  typeTable.setType(d, typeTable.getVoidType());

  return true;
}

bool SemanticAnalysis::visit(NewObject *n) {
  auto type = typeTable.getType(n->getType());

  if (!type->getClass()) {
    throw WrongTypeException(type, n);
  }

  typeTable.setType(n, type);
  return true;
}

bool SemanticAnalysis::visit(IfStatement *i) {
	auto condExpr = i->getCond();
	auto thenExpr = i->getThen();
	auto elseExpr = i->getElse();

	// TODO: Create new symbol table scopes, check type compatibility, return the right type
	// throw WrongTypeException if anything wrong

	if (!condExpr) {
		throw MissingIfCondException(i);
	}
	if (!visit(condExpr)) {
		return false;
	}

	if (!thenExpr) {
		throw MissingIfThenException(i);
	}
	if (!visit(thenExpr)) {
		return false;
	}

	if (elseExpr && !visit(elseExpr)) {
		return false;
	}

	typeTable.setType(i, typeTable.getVoidType());

  return true;
}

bool SemanticAnalysis::visit(WhileStatement *w) {
  auto cond = w->getCond();

  if (!cond) {
    throw MissingWhileCondException(w);
  }

  if (!visit(cond)) {
    return false;
  }

  if (typeTable.getType(cond) != typeTable.getIntType()) {
    throw WrongTypeException(typeTable.getType(cond), typeTable.getIntType(),
                             cond);
  }

  auto body = w->getBody();
  if (body) {
    SymbolTable::Scope whileScope(symbolTable);
    if (!visit(body)) {
      return false;
    }
  }

  typeTable.setType(w, typeTable.getVoidType());

  return true;
}

bool SemanticAnalysis::visit(LocalDefinition *local) {
  symbolTable.insert(local);
  typeTable.setType(local, typeTable.getType(local->getType()));

  if (local->getInit()) {
    if (!visit(local->getInit())) {
      return false;
    }
  }

  return true;
}
